@@ -7,7 +7,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/commands/copy.c,v 1.136.2.1 2001/08/08 22:32:29 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/commands/copy.c,v 1.145.2.1 2002/02/27 01:47:31 ishii Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -41,7 +41,7 @@
 #endif
 
 #define ISOCTAL(c) (((c) >= '0') && ((c) <= '7'))
-#define VALUE(c) ((c) - '0')
+#define OCTVALUE(c) ((c) - '0')
 
 
 /* non-export function prototypes */
@@ -59,7 +59,7 @@ static const char BinarySignature[12] = "PGBCOPY\n\377\r\n\0";
  * Static communication variables ... pretty grotty, but COPY has
  * never been reentrant...
  */
-int			lineno = 0;			/* exported for use by elog() -- dz */
+int			copy_lineno = 0;	/* exported for use by elog() -- dz */
 static bool fe_eof;
 
 /*
@@ -76,7 +76,6 @@ static StringInfoData attribute_buf;
 #ifdef MULTIBYTE
 static int	client_encoding;
 static int	server_encoding;
-
 #endif
 
 
@@ -84,13 +83,13 @@ static int	server_encoding;
  * Internal communications functions
  */
 static void CopySendData(void *databuf, int datasize, FILE *fp);
-static void CopySendString(char *str, FILE *fp);
+static void CopySendString(const char *str, FILE *fp);
 static void CopySendChar(char c, FILE *fp);
 static void CopyGetData(void *databuf, int datasize, FILE *fp);
 static int	CopyGetChar(FILE *fp);
 static int	CopyGetEof(FILE *fp);
 static int	CopyPeekChar(FILE *fp);
-static void CopyDonePeek(FILE *fp, int c, int pickup);
+static void CopyDonePeek(FILE *fp, int c, bool pickup);
 
 /*
  * CopySendData sends output data either to the file
@@ -119,9 +118,9 @@ CopySendData(void *databuf, int datasize, FILE *fp)
 }
 
 static void
-CopySendString(char *str, FILE *fp)
+CopySendString(const char *str, FILE *fp)
 {
-	CopySendData(str, strlen(str), fp);
+	CopySendData((void *) str, strlen(str), fp);
 }
 
 static void
@@ -158,13 +157,10 @@ CopyGetChar(FILE *fp)
 {
 	if (!fp)
 	{
-		unsigned char ch;
+		int			ch = pq_getbyte();
 
-		if (pq_getbytes((char *) &ch, 1))
-		{
+		if (ch == EOF)
 			fe_eof = true;
-			return EOF;
-		}
 		return ch;
 	}
 	else
@@ -182,10 +178,12 @@ CopyGetEof(FILE *fp)
 
 /*
  * CopyPeekChar reads a byte in "peekable" mode.
+ *
  * after each call to CopyPeekChar, a call to CopyDonePeek _must_
  * follow, unless EOF was returned.
- * CopyDonePeek will either take the peeked char off the steam
- * (if pickup is != 0) or leave it on the stream (if pickup == 0)
+ *
+ * CopyDonePeek will either take the peeked char off the stream
+ * (if pickup is true) or leave it on the stream (if pickup is false).
  */
 static int
 CopyPeekChar(FILE *fp)
@@ -203,20 +201,14 @@ CopyPeekChar(FILE *fp)
 }
 
 static void
-CopyDonePeek(FILE *fp, int c, int pickup)
+CopyDonePeek(FILE *fp, int c, bool pickup)
 {
 	if (!fp)
 	{
 		if (pickup)
 		{
-
-			/*
-			 * We want to pick it up - just receive again into dummy
-			 * buffer
-			 */
-			char		c;
-
-			pq_getbytes(&c, 1);
+			/* We want to pick it up */
+			(void) pq_getbyte();
 		}
 		/* If we didn't want to pick it up, just leave it where it sits */
 	}
@@ -227,7 +219,7 @@ CopyDonePeek(FILE *fp, int c, int pickup)
 			/* We don't want to pick it up - so put it back in there */
 			ungetc(c, fp);
 		}
-		/* If we wanted to pick it up, it's already there */
+		/* If we wanted to pick it up, it's already done */
 	}
 }
 
@@ -235,43 +227,45 @@ CopyDonePeek(FILE *fp, int c, int pickup)
 
 /*
  *	 DoCopy executes the SQL COPY statement.
+ *
+ * Either unload or reload contents of table <relname>, depending on <from>.
+ * (<from> = TRUE means we are inserting into the table.)
+ *
+ * If <pipe> is false, transfer is between the table and the file named
+ * <filename>.	Otherwise, transfer is between the table and our regular
+ * input/output stream. The latter could be either stdin/stdout or a
+ * socket, depending on whether we're running under Postmaster control.
+ *
+ * Iff <binary>, unload or reload in the binary format, as opposed to the
+ * more wasteful but more robust and portable text format.
+ *
+ * Iff <oids>, unload or reload the format that includes OID information.
+ * On input, we accept OIDs whether or not the table has an OID column,
+ * but silently drop them if it does not.  On output, we report an error
+ * if the user asks for OIDs in a table that has none (not providing an
+ * OID column might seem friendlier, but could seriously confuse programs).
+ *
+ * If in the text format, delimit columns with delimiter <delim> and print
+ * NULL values as <null_print>.
+ *
+ * When loading in the text format from an input stream (as opposed to
+ * a file), recognize a "." on a line by itself as EOF. Also recognize
+ * a stream EOF.  When unloading in the text format to an output stream,
+ * write a "." on a line by itself at the end of the data.
+ *
+ * Do not allow a Postgres user without superuser privilege to read from
+ * or write to a file.
+ *
+ * Do not allow the copy if user doesn't have proper permission to access
+ * the table.
  */
-
 void
 DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 	   char *filename, char *delim, char *null_print)
 {
-/*----------------------------------------------------------------------------
-  Either unload or reload contents of class <relname>, depending on <from>.
-
-  If <pipe> is false, transfer is between the class and the file named
-  <filename>.  Otherwise, transfer is between the class and our regular
-  input/output stream.	The latter could be either stdin/stdout or a
-  socket, depending on whether we're running under Postmaster control.
-
-  Iff <binary>, unload or reload in the binary format, as opposed to the
-  more wasteful but more robust and portable text format.
-
-  If in the text format, delimit columns with delimiter <delim> and print
-  NULL values as <null_print>.
-
-  When loading in the text format from an input stream (as opposed to
-  a file), recognize a "." on a line by itself as EOF.	Also recognize
-  a stream EOF.  When unloading in the text format to an output stream,
-  write a "." on a line by itself at the end of the data.
-
-  Iff <oids>, unload or reload the format that includes OID information.
-
-  Do not allow a Postgres user without superuser privilege to read from
-  or write to a file.
-
-  Do not allow the copy if user doesn't have proper permission to access
-  the class.
-----------------------------------------------------------------------------*/
-
 	FILE	   *fp;
 	Relation	rel;
-	const AclMode required_access = from ? ACL_WR : ACL_RD;
+	const AclMode required_access = (from ? ACL_INSERT : ACL_SELECT);
 	int			result;
 
 	/*
@@ -294,6 +288,12 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 	if (pipe && binary)
 		elog(ERROR, "COPY BINARY is not supported to stdout or from stdin");
 
+	/*
+	 * Presently, only single-character delimiter strings are supported.
+	 */
+	if (strlen(delim) != 1)
+		elog(ERROR, "COPY delimiter must be a single character");
+
 	/*
 	 * Set up variables to avoid per-attribute overhead.
 	 */
@@ -422,6 +422,10 @@ CopyTo(Relation rel, bool binary, bool oids, FILE *fp,
 	int16		fld_size;
 	char	   *string;
 
+	if (oids && !rel->rd_rel->relhasoids)
+		elog(ERROR, "COPY: table %s does not have OIDs",
+			 RelationGetRelationName(rel));
+
 	tupDesc = rel->rd_att;
 	attr_count = rel->rd_att->natts;
 	attr = rel->rd_att->attrs;
@@ -528,7 +532,6 @@ CopyTo(Relation rel, bool binary, bool oids, FILE *fp,
 			}
 			else
 			{
-
 				/*
 				 * If we have a toasted datum, forcibly detoast it to
 				 * avoid memory leakage inside the type's output routine
@@ -652,6 +655,7 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
 	resultRelInfo = makeNode(ResultRelInfo);
 	resultRelInfo->ri_RangeTableIndex = 1;		/* dummy */
 	resultRelInfo->ri_RelationDesc = rel;
+	resultRelInfo->ri_TrigDesc = rel->trigdesc;
 
 	ExecOpenIndices(resultRelInfo);
 
@@ -717,17 +721,21 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
 		elements = NULL;
 	}
 
+	/* Silently drop incoming OIDs if table does not have OIDs */
+	if (!rel->rd_rel->relhasoids)
+		oids = false;
+
 	values = (Datum *) palloc(attr_count * sizeof(Datum));
 	nulls = (char *) palloc(attr_count * sizeof(char));
 
-	lineno = 0;
+	copy_lineno = 0;
 	fe_eof = false;
 
 	while (!done)
 	{
 		CHECK_FOR_INTERRUPTS();
 
-		lineno++;
+		copy_lineno++;
 
 		/* Reset the per-output-tuple exprcontext */
 		ResetPerTupleExprContext(estate);
@@ -884,12 +892,12 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
 		skip_tuple = false;
 
 		/* BEFORE ROW INSERT Triggers */
-		if (rel->trigdesc &&
-			rel->trigdesc->n_before_row[TRIGGER_EVENT_INSERT] > 0)
+		if (resultRelInfo->ri_TrigDesc &&
+			resultRelInfo->ri_TrigDesc->n_before_row[TRIGGER_EVENT_INSERT] > 0)
 		{
 			HeapTuple	newtuple;
 
-			newtuple = ExecBRInsertTriggers(estate, rel, tuple);
+			newtuple = ExecBRInsertTriggers(estate, resultRelInfo, tuple);
 
 			if (newtuple == NULL)		/* "do nothing" */
 				skip_tuple = true;
@@ -919,8 +927,8 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
 				ExecInsertIndexTuples(slot, &(tuple->t_self), estate, false);
 
 			/* AFTER ROW INSERT Triggers */
-			if (rel->trigdesc)
-				ExecARInsertTriggers(estate, rel, tuple);
+			if (resultRelInfo->ri_TrigDesc)
+				ExecARInsertTriggers(estate, resultRelInfo, tuple);
 		}
 
 		for (i = 0; i < attr_count; i++)
@@ -935,7 +943,7 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
 	/*
 	 * Done, clean up
 	 */
-	lineno = 0;
+	copy_lineno = 0;
 
 	pfree(values);
 	pfree(nulls);
@@ -1007,7 +1015,7 @@ CopyReadNewline(FILE *fp, int *newline)
  * Note that the caller should not pfree the string!
  *
  * *isnull is set true if a null attribute, else false.
- * delim is the string of acceptable delimiter characters(s).
+ * delim is the column delimiter string (currently always 1 character).
  * *newline remembers whether we've seen a newline ending this tuple.
  * null_print says how NULL values are represented
  */
@@ -1016,6 +1024,7 @@ static char *
 CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_print)
 {
 	int			c;
+	int			delimc = (unsigned char)delim[0];
 
 #ifdef MULTIBYTE
 	int			mblen;
@@ -1049,7 +1058,7 @@ CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_
 			*newline = 1;
 			break;
 		}
-		if (strchr(delim, c))
+		if (c == delimc)
 			break;
 		if (c == '\\')
 		{
@@ -1069,31 +1078,30 @@ CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_
 					{
 						int			val;
 
-						val = VALUE(c);
+						val = OCTVALUE(c);
 						c = CopyPeekChar(fp);
 						if (ISOCTAL(c))
 						{
-							val = (val << 3) + VALUE(c);
-							CopyDonePeek(fp, c, 1);		/* Pick up the
-														 * character! */
+							val = (val << 3) + OCTVALUE(c);
+							CopyDonePeek(fp, c, true /*pick up*/);
 							c = CopyPeekChar(fp);
 							if (ISOCTAL(c))
 							{
-								CopyDonePeek(fp, c, 1); /* pick up! */
-								val = (val << 3) + VALUE(c);
+								val = (val << 3) + OCTVALUE(c);
+								CopyDonePeek(fp, c, true /*pick up*/);
 							}
 							else
 							{
 								if (c == EOF)
 									goto endOfFile;
-								CopyDonePeek(fp, c, 0); /* Return to stream! */
+								CopyDonePeek(fp, c, false /*put back*/);
 							}
 						}
 						else
 						{
 							if (c == EOF)
 								goto endOfFile;
-							CopyDonePeek(fp, c, 0);		/* Return to stream! */
+							CopyDonePeek(fp, c, false /*put back*/);
 						}
 						c = val & 0377;
 					}
@@ -1135,6 +1143,7 @@ CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_
 		}
 		appendStringInfoCharMacro(&attribute_buf, c);
 #ifdef MULTIBYTE
+		/* XXX shouldn't this be done even when encoding is the same? */
 		if (client_encoding != server_encoding)
 		{
 			/* get additional bytes of the char, if any */
@@ -1181,16 +1190,18 @@ CopyAttributeOut(FILE *fp, char *server_string, char *delim)
 {
 	char	   *string;
 	char		c;
+	char		delimc = delim[0];
 
 #ifdef MULTIBYTE
+	bool		same_encoding;
 	char	   *string_start;
 	int			mblen;
 	int			i;
-
 #endif
 
 #ifdef MULTIBYTE
-	if (client_encoding != server_encoding)
+	same_encoding = (server_encoding == client_encoding);
+	if (!same_encoding)
 	{
 		string = (char *) pg_server_to_client((unsigned char *) server_string,
 											  strlen(server_string));
@@ -1199,31 +1210,64 @@ CopyAttributeOut(FILE *fp, char *server_string, char *delim)
 	else
 	{
 		string = server_string;
-		string_start = NULL;	/* unused, but keep compiler quiet */
+		string_start = NULL;
 	}
 #else
 	string = server_string;
 #endif
 
 #ifdef MULTIBYTE
-	for (; (mblen = (server_encoding == client_encoding ? 1 : pg_encoding_mblen(client_encoding, string))) &&
-		 ((c = *string) != '\0'); string += mblen)
+	for (; (c = *string) != '\0'; string += mblen)
 #else
 	for (; (c = *string) != '\0'; string++)
 #endif
 	{
-		if (c == delim[0] || c == '\n' || c == '\\')
-			CopySendChar('\\', fp);
 #ifdef MULTIBYTE
-		for (i = 0; i < mblen; i++)
-			CopySendChar(*(string + i), fp);
-#else
-		CopySendChar(c, fp);
+		mblen = 1;
 #endif
+		switch (c)
+		{
+			case '\b':
+				CopySendString("\\b", fp);
+				break;
+			case '\f':
+				CopySendString("\\f", fp);
+				break;
+			case '\n':
+				CopySendString("\\n", fp);
+				break;
+			case '\r':
+				CopySendString("\\r", fp);
+				break;
+			case '\t':
+				CopySendString("\\t", fp);
+				break;
+			case '\v':
+				CopySendString("\\v", fp);
+				break;
+			case '\\':
+				CopySendString("\\\\", fp);
+				break;
+			default:
+				if (c == delimc)
+					CopySendChar('\\', fp);
+				CopySendChar(c, fp);
+#ifdef MULTIBYTE
+				/* XXX shouldn't this be done even when encoding is same? */
+				if (!same_encoding)
+				{
+					/* send additional bytes of the char, if any */
+					mblen = pg_encoding_mblen(client_encoding, string);
+					for (i = 1; i < mblen; i++)
+						CopySendChar(string[i], fp);
+				}
+#endif
+				break;
+		}
 	}
 
 #ifdef MULTIBYTE
-	if (client_encoding != server_encoding)
+	if (string_start)
 		pfree(string_start);	/* pfree pg_server_to_client result */
 #endif
 }
